from __future__ import annotations

import datetime as dt
from pathlib import Path

from click.testing import CliRunner
from freezegun import freeze_time
from hypothesis import given
from hypothesis.strategies import integers

from utilities.datetime import TODAY_UTC
from utilities.hypothesis import temp_paths
from utilities.pathvalidate import valid_path
from utilities.pytest import skipif_windows
from utilities.scripts.clean_dir import main
from utilities.scripts.clean_dir.classes import Config


class TestCleanDir:
    timedelta = dt.timedelta(days=Config().days + 1)

    @skipif_windows
    def test_file(self, *, tmp_path: Path) -> None:
        valid_path(tmp_path, "file").touch()
        runner = CliRunner()
        args = ["--path", str(tmp_path)]
        with freeze_time(TODAY_UTC + self.timedelta):
            result = runner.invoke(main, args)
        assert result.exit_code == 0

    @skipif_windows
    def test_dir_to_remove(self, *, tmp_path: Path) -> None:
        valid_path(tmp_path, "dir").mkdir()
        runner = CliRunner()
        args = ["--path", str(tmp_path)]
        result = runner.invoke(main, args)
        assert result.exit_code == 0

    @skipif_windows
    def test_dir_to_retain(self, *, tmp_path: Path) -> None:
        dir_ = valid_path(tmp_path, "dir")
        dir_.mkdir()
        valid_path(dir_, "file").touch()
        runner = CliRunner()
        args = ["--path", str(tmp_path)]
        result = runner.invoke(main, args)
        assert result.exit_code == 0

    @skipif_windows
    def test_symlink(self, *, tmp_path: Path) -> None:
        file = valid_path(tmp_path, "file")
        file.touch()
        valid_path(tmp_path, "second").symlink_to(file)
        runner = CliRunner()
        args = ["--path", str(tmp_path)]
        with freeze_time(TODAY_UTC + self.timedelta):
            result = runner.invoke(main, args)
        assert result.exit_code == 0

    @skipif_windows
    @given(root=temp_paths(), chunk_size=integers(1, 10))
    def test_chunk_size(self, *, root: Path, chunk_size: int) -> None:
        valid_path(root, "file").touch()
        runner = CliRunner()
        args = ["--path", str(root), "--chunk-size", str(chunk_size)]
        with freeze_time(TODAY_UTC + self.timedelta):
            result = runner.invoke(main, args)
        assert result.exit_code == 0

    @skipif_windows
    def test_dry_run(self, *, tmp_path: Path) -> None:
        valid_path(tmp_path, "file").touch()
        runner = CliRunner()
        args = ["--path", str(tmp_path), "--dry-run"]
        with freeze_time(TODAY_UTC + self.timedelta):
            result = runner.invoke(main, args)
        assert result.exit_code == 0



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from functools import reduce
from itertools import chain
from typing import Any

from polars import Boolean, DataFrame, Expr, PolarsDataType, col, lit, when
from polars.exceptions import ColumnNotFoundError, OutOfBoundsError
from polars.testing import assert_frame_equal
from polars.type_aliases import IntoExpr, JoinStrategy, JoinValidation, SchemaDict

from utilities.errors import redirect_error
from utilities.math import is_equal_or_approx
from utilities.types import SequenceStrs


def check_polars_dataframe(
    df: DataFrame,
    /,
    *,
    columns: SequenceStrs | None = None,
    dtypes: list[PolarsDataType] | None = None,
    height: int | tuple[int, float] | None = None,
    min_height: int | None = None,
    max_height: int | None = None,
    predicates: Mapping[str, Callable[[Any], bool]] | None = None,
    schema: SchemaDict | None = None,
    schema_inc: SchemaDict | None = None,
    shape: tuple[int, int] | None = None,
    sorted: IntoExpr | Iterable[IntoExpr] | None = None,  # noqa: A002
    unique: IntoExpr | Iterable[IntoExpr] | None = None,
    width: int | None = None,
) -> None:
    """Check the properties of a DataFrame."""
    if (columns is not None) and (df.columns != list(columns)):
        msg = f"{df=}, {columns=}"
        raise CheckPolarsDataFrameError(msg)
    if (dtypes is not None) and (df.dtypes != dtypes):
        msg = f"{df=}, {dtypes=}"
        raise CheckPolarsDataFrameError(msg)
    if (height is not None) and not is_equal_or_approx(df.height, height):
        msg = f"{df=}, {height=}"
        raise CheckPolarsDataFrameError(msg)
    if (min_height is not None) and (len(df) < min_height):
        msg = f"{df=}, {min_height=}"
        raise CheckPolarsDataFrameError(msg)
    if (max_height is not None) and (len(df) > max_height):
        msg = f"{df=}, {max_height=}"
        raise CheckPolarsDataFrameError(msg)
    if predicates is not None:
        _check_polars_dataframe_predicates(df, predicates)
    if schema is not None:
        _check_polars_dataframe_schema(df, schema)
    if schema_inc is not None:
        _check_polars_dataframe_schema_inc(df, schema_inc)
    if (shape is not None) and (df.shape != shape):
        msg = f"{df=}"
        raise CheckPolarsDataFrameError(msg)
    if sorted is not None:
        df_sorted = df.sort(sorted)
        with redirect_error(AssertionError, CheckPolarsDataFrameError(f"{df=}")):
            assert_frame_equal(df, df_sorted)
    if (unique is not None) and df.select(unique).is_duplicated().any():
        msg = f"{df=}, {unique=}"
        raise CheckPolarsDataFrameError(msg)
    if (width is not None) and (df.width != width):
        msg = f"{df=}"
        raise CheckPolarsDataFrameError(msg)


def _check_polars_dataframe_predicates(
    df: DataFrame, predicates: Mapping[str, Callable[[Any], bool]], /
) -> None:
    missing: set[str] = set()
    failed: set[str] = set()
    for column, predicate in predicates.items():
        try:
            sr = df[column]
        except ColumnNotFoundError:  # noqa: PERF203
            missing.add(column)
        else:
            if not sr.map_elements(predicate, return_dtype=Boolean).all():
                failed.add(column)
    if (len(missing) >= 1) or (len(failed)) >= 1:
        msg = f"{missing=}, {failed=}"
        raise CheckPolarsDataFrameError(msg)


def _check_polars_dataframe_schema(df: DataFrame, schema: SchemaDict, /) -> None:
    if df.schema != schema:
        set_act, set_exp = map(set, [df.schema, schema])
        extra = set_act - set_exp
        missing = set_exp - set_act
        differ = {
            col: (left, right)
            for col in set_act & set_exp
            if (left := df.schema[col]) != (right := schema[col])
        }
        msg = f"{df=}, {extra=}, {missing=}, {differ=}"
        raise CheckPolarsDataFrameError(msg)


def _check_polars_dataframe_schema_inc(
    df: DataFrame, schema_inc: SchemaDict, /
) -> None:
    missing: set[str] = set()
    wrong_dtype: set[str] = set()
    for column, dtype in schema_inc.items():
        try:
            sr = df[column]
        except ColumnNotFoundError:  # noqa: PERF203
            missing.add(column)
        else:
            if sr.dtype != dtype:
                wrong_dtype.add(column)
    if (len(missing) >= 1) or (len(wrong_dtype)) >= 1:
        msg = f"{missing=}, {wrong_dtype=}"
        raise CheckPolarsDataFrameError(msg)


class CheckPolarsDataFrameError(Exception):
    ...


def join(
    df: DataFrame,
    *dfs: DataFrame,
    on: str | Expr | Sequence[str | Expr],
    how: JoinStrategy = "inner",
    validate: JoinValidation = "m:m",
) -> DataFrame:
    def inner(left: DataFrame, right: DataFrame, /) -> DataFrame:
        return left.join(right, on=on, how=how, validate=validate)

    return reduce(inner, chain([df], dfs))


def nan_sum_agg(column: str | Expr, /, *, dtype: PolarsDataType | None = None) -> Expr:
    """Nan sum aggregation."""

    col_use = col(column) if isinstance(column, str) else column
    return (
        when(col_use.is_not_null().any())
        .then(col_use.sum())
        .otherwise(lit(None, dtype=dtype))
    )


def nan_sum_cols(
    column: str | Expr, *columns: str | Expr, dtype: PolarsDataType | None = None
) -> Expr:
    """Nan sum across columns."""

    all_columns = chain([column], columns)
    all_exprs = (
        col(column) if isinstance(column, str) else column for column in all_columns
    )

    def func(x: Expr, y: Expr, /) -> Expr:
        return (
            when(x.is_not_null() & y.is_not_null())
            .then(x + y)
            .when(x.is_not_null() & y.is_null())
            .then(x)
            .when(x.is_null() & y.is_not_null())
            .then(y)
            .otherwise(lit(None, dtype=dtype))
        )

    return reduce(func, all_exprs)


@contextmanager
def redirect_empty_polars_concat() -> Iterator[None]:
    """Redirect to the `EmptyPolarsConcatError`."""
    with redirect_error(
        ValueError, EmptyPolarsConcatError, match="cannot concat empty list"
    ):
        yield


class EmptyPolarsConcatError(Exception):
    ...


def set_first_row_as_columns(df: DataFrame, /) -> DataFrame:
    """Set the first row of a DataFrame as its columns."""

    with redirect_error(OutOfBoundsError, SetFirstRowAsColumnsError(f"{df=}")):
        row = df.row(0)
    mapping = dict(zip(df.columns, row, strict=True))
    return df[1:].rename(mapping)


class SetFirstRowAsColumnsError(Exception):
    ...


__all__ = [
    "CheckPolarsDataFrameError",
    "EmptyPolarsConcatError",
    "SetFirstRowAsColumnsError",
    "check_polars_dataframe",
    "join",
    "nan_sum_agg",
    "nan_sum_cols",
    "redirect_empty_polars_concat",
    "set_first_row_as_columns",
]


try:
    from utilities._polars.bs4 import (
        TableTagToDataFrameError,
        table_tag_to_dataframe,
        yield_tables,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["TableTagToDataFrameError", "table_tag_to_dataframe", "yield_tables"]



from __future__ import annotations

from collections.abc import Iterator

from hypothesis import given
from hypothesis.strategies import binary, dictionaries, integers, lists, text
from pytest import raises

from utilities.more_itertools import (
    OneEmptyError,
    OneNonUniqueError,
    always_iterable,
    one,
)


class TestAlwaysIterable:
    @given(x=binary())
    def test_bytes(self, *, x: bytes) -> None:
        assert list(always_iterable(x)) == [x]
        assert list(always_iterable(x, base_type=None)) == list(x)
        assert list(always_iterable(x, base_type=bytes)) == [x]
        assert list(always_iterable(x, base_type=(bytes,))) == [x]

    @given(x=integers())
    def test_integer(self, *, x: int) -> None:
        assert list(always_iterable(x)) == [x]
        assert list(always_iterable(x, base_type=None)) == [x]
        assert list(always_iterable(x, base_type=int)) == [x]
        assert list(always_iterable(x, base_type=(int,))) == [x]

    @given(x=text())
    def test_string(self, *, x: str) -> None:
        assert list(always_iterable(x)) == [x]
        assert list(always_iterable(x, base_type=None)) == list(x)
        assert list(always_iterable(x, base_type=str)) == [x]
        assert list(always_iterable(x, base_type=(str,))) == [x]

    @given(x=dictionaries(text(), integers()))
    def test_dict(self, *, x: dict[str, int]) -> None:
        assert list(always_iterable(x)) == list(x)
        assert list(always_iterable(x, base_type=dict)) == [x]
        assert list(always_iterable(x, base_type=(dict,))) == [x]

    @given(x=lists(integers()))
    def test_lists(self, *, x: list[int]) -> None:
        assert list(always_iterable(x)) == x
        assert list(always_iterable(x, base_type=None)) == x
        assert list(always_iterable(x, base_type=list)) == [x]
        assert list(always_iterable(x, base_type=(list,))) == [x]

    def test_none(self) -> None:
        assert list(always_iterable(None)) == []

    def test_generator(self) -> None:
        def yield_ints() -> Iterator[int]:
            yield 0
            yield 1

        assert list(always_iterable(yield_ints())) == [0, 1]


class TestOne:
    def test_main(self) -> None:
        assert one([None]) is None

    def test_error_empty(self) -> None:
        with raises(
            OneEmptyError,
            match="Iterable .* must contain exactly one item; it was empty",
        ):
            _ = one([])

    def test_error_non_unique(self) -> None:
        with raises(
            OneNonUniqueError,
            match="Iterable .* must contain exactly one item; got .*, .* and perhaps more",
        ):
            _ = one([1, 2])



from __future__ import annotations

import datetime as dt
from collections.abc import Iterator
from contextlib import suppress
from datetime import tzinfo
from re import sub
from typing import cast

from typing_extensions import Never, assert_never

from utilities.errors import redirect_error
from utilities.platform import SYSTEM, System
from utilities.re import ExtractGroupsError, extract_groups
from utilities.types import Duration
from utilities.zoneinfo import HONG_KONG, TOKYO

UTC = dt.timezone.utc
EPOCH_UTC = dt.datetime.fromtimestamp(0, tz=UTC)


def add_weekdays(date: dt.date, /, *, n: int = 1) -> dt.date:
    """Add a number of a weekdays to a given date.

    If the initial date is a weekend, then moving to the adjacent weekday
    counts as 1 move.
    """
    if n == 0 and not is_weekday(date):
        raise AddWeekdaysError(date)
    if n >= 1:
        for _ in range(n):
            date = round_to_next_weekday(date + dt.timedelta(days=1))
    elif n <= -1:
        for _ in range(-n):
            date = round_to_prev_weekday(date - dt.timedelta(days=1))
    return date


class AddWeekdaysError(Exception):
    ...


def date_to_datetime(
    date: dt.date, /, *, time: dt.time = dt.time(0), tzinfo: tzinfo | None = UTC
) -> dt.datetime:
    """Expand a date into a datetime."""
    return dt.datetime.combine(date, time, tzinfo=tzinfo)


def duration_to_float(duration: Duration, /) -> float:
    """Ensure the duration is a float."""
    if isinstance(duration, int):
        return float(duration)
    if isinstance(duration, float):
        return duration
    return duration.total_seconds()


def duration_to_timedelta(duration: Duration, /) -> dt.timedelta:
    """Ensure the duration is a timedelta."""
    if isinstance(duration, int | float):
        return dt.timedelta(seconds=duration)
    return duration


def ensure_date(date: dt.date | str, /) -> dt.date:
    """Ensure the object is a date."""
    return date if isinstance(date, dt.date) else parse_date(date)


def ensure_datetime(datetime: dt.datetime | str, /) -> dt.datetime:
    """Ensure the object is a datetime."""
    if isinstance(datetime, dt.datetime):
        return datetime
    return parse_datetime(datetime)


def ensure_time(time: dt.time | str, /) -> dt.time:
    """Ensure the object is a time."""
    return time if isinstance(time, dt.time) else parse_time(time)


def ensure_timedelta(timedelta: dt.timedelta | str, /) -> dt.timedelta:
    """Ensure the object is a timedelta."""
    if isinstance(timedelta, dt.timedelta):
        return timedelta
    return parse_timedelta(timedelta)


def get_now(*, tz: tzinfo | None = UTC) -> dt.datetime:
    """Get the current, timezone-aware time."""
    return dt.datetime.now(tz=tz)


NOW_UTC, NOW_HKG, NOW_TKY = (get_now(tz=tz) for tz in [UTC, HONG_KONG, TOKYO])


def get_today(*, tz: tzinfo | None = UTC) -> dt.date:
    """Get the current, timezone-aware date."""
    return get_now(tz=tz).date()


TODAY_UTC, TODAY_HKG, TODAY_TKY = (get_today(tz=tz) for tz in [UTC, HONG_KONG, TOKYO])


def is_equal_mod_tz(x: dt.datetime, y: dt.datetime, /) -> bool:
    """Check if x == y, modulo timezone."""
    x_aware, y_aware = x.tzinfo is not None, y.tzinfo is not None
    match x_aware, y_aware:
        case (False, False) | (True, True):
            return x == y
        case True, False:
            return x.astimezone(UTC).replace(tzinfo=None) == y
        case False, True:
            return x == y.astimezone(UTC).replace(tzinfo=None)
        case _ as never:
            assert_never(cast(Never, never))


def is_weekday(date: dt.date, /) -> bool:
    """Check if a date is a weekday."""
    friday = 5
    return date.isoweekday() <= friday


def local_timezone() -> tzinfo:
    """Get the local timezone."""
    tz = get_now(tz=None).astimezone().tzinfo
    if tz is None:  # pragma: no cover
        msg = f"{tz=}"
        raise LocalTimeZoneError(msg)
    return tz


class LocalTimeZoneError(Exception):
    ...


def maybe_sub_pct_y(text: str, /) -> str:
    """Substitute the `%Y' token with '%4Y' if necessary."""
    match SYSTEM:
        case System.windows:  # pragma: os-ne-windows
            return text
        case System.mac:  # pragma: os-ne-macos
            return text
        case System.linux:  # pragma: os-ne-linux
            return sub("%Y", "%4Y", text)
        case _ as never:  # type: ignore
            assert_never(never)


def parse_date(date: str, /, *, tzinfo: tzinfo = UTC) -> dt.date:
    """Parse a string into a date."""
    with suppress(ValueError):
        return dt.date.fromisoformat(date)
    for fmt in ["%Y%m%d", "%Y %m %d", "%d%b%Y", "%d %b %Y"]:
        with suppress(ValueError):  # pragma: version-ge-311
            return dt.datetime.strptime(date, fmt).replace(tzinfo=tzinfo).date()
    raise ParseDateError(date)


class ParseDateError(Exception):
    ...


def parse_datetime(datetime: str, /, *, tzinfo: tzinfo = UTC) -> dt.datetime:
    """Parse a string into a datetime."""
    with suppress(ValueError):
        return dt.datetime.fromisoformat(datetime).replace(tzinfo=UTC)
    for fmt in [
        "%Y%m%d",
        "%Y%m%dT%H",
        "%Y%m%dT%H%M",
        "%Y%m%dT%H%M%S",
        "%Y%m%dT%H%M%S.%f",
    ]:
        with suppress(ValueError):  # pragma: version-ge-311
            return dt.datetime.strptime(datetime, fmt).replace(tzinfo=tzinfo)
    for fmt in ["%Y-%m-%d %H:%M:%S.%f%z", "%Y%m%dT%H%M%S.%f%z"]:
        with suppress(ValueError):  # pragma: version-ge-311
            return dt.datetime.strptime(datetime, fmt)  # noqa: DTZ007
    raise ParseDateTimeError(datetime)


class ParseDateTimeError(Exception):
    ...


def parse_time(time: str, /) -> dt.time:
    """Parse a string into a time."""
    with suppress(ValueError):
        return dt.time.fromisoformat(time)
    for fmt in ["%H", "%H%M", "%H%M%S", "%H%M%S.%f"]:  # pragma: version-ge-311
        with suppress(ValueError):
            return dt.datetime.strptime(time, fmt).replace(tzinfo=UTC).time()
    raise ParseTimeError(time)


class ParseTimeError(Exception):
    ...


def parse_timedelta(timedelta: str, /) -> dt.timedelta:
    """Parse a string into a timedelta."""

    def try_parse(fmt: str, /) -> dt.datetime | None:
        try:
            return dt.datetime.strptime(timedelta, fmt).replace(tzinfo=UTC)
        except ValueError:
            return None

    try:
        as_dt = next(
            parsed
            for fmt in ("%H:%M:%S", "%H:%M:%S.%f")
            if (parsed := try_parse(fmt)) is not None
        )
    except StopIteration:
        pass
    else:
        return dt.timedelta(
            hours=as_dt.hour,
            minutes=as_dt.minute,
            seconds=as_dt.second,
            microseconds=as_dt.microsecond,
        )
    with redirect_error(ExtractGroupsError, ParseTimedeltaError(f"{timedelta=}")):
        days, tail = extract_groups(r"([-\d]+)\s*(?:days?)?,?\s*([\d:\.]+)", timedelta)
    return dt.timedelta(days=int(days)) + parse_timedelta(tail)


class ParseTimedeltaError(Exception):
    ...


def round_to_next_weekday(date: dt.date, /) -> dt.date:
    """Round a date to the next weekday."""
    return _round_to_weekday(date, is_next=True)


def round_to_prev_weekday(date: dt.date, /) -> dt.date:
    """Round a date to the previous weekday."""
    return _round_to_weekday(date, is_next=False)


def _round_to_weekday(date: dt.date, /, *, is_next: bool) -> dt.date:
    """Round a date to the previous weekday."""
    n = 1 if is_next else -1
    while not is_weekday(date):
        date = add_weekdays(date, n=n)
    return date


def serialize_date(date: dt.date, /) -> str:
    """Serialize a date."""
    if isinstance(date, dt.datetime):
        return serialize_date(date.date())
    return date.isoformat()


def serialize_datetime(datetime: dt.datetime, /) -> str:
    """Serialize a datetime."""
    return datetime.isoformat()


def serialize_time(time: dt.time, /) -> str:
    """Serialize a time."""
    return time.isoformat()


def serialize_timedelta(timedelta: dt.timedelta, /) -> str:
    """Serialize a timedelta."""
    if (days := timedelta.days) == 0:
        return str(timedelta)
    tail = serialize_timedelta(timedelta - dt.timedelta(days=days))
    return f"d{days},{tail}"


def yield_weekdays(
    *, start: dt.date | None = None, end: dt.date | None = None, days: int | None = None
) -> Iterator[dt.date]:
    """Yield the weekdays in a range."""
    if (start is not None) and (end is not None) and (days is None):
        date = round_to_next_weekday(start)
        while date <= end:
            yield date
            date = round_to_next_weekday(date + dt.timedelta(days=1))
    elif (start is not None) and (end is None) and (days is not None):
        date = round_to_next_weekday(start)
        for _ in range(days):
            yield date
            date = round_to_next_weekday(date + dt.timedelta(days=1))
    elif (start is None) and (end is not None) and (days is not None):
        date = round_to_prev_weekday(end)
        for _ in range(days):
            yield date
            date = round_to_prev_weekday(date - dt.timedelta(days=1))
    else:
        msg = f"Invalid arguments: {start=}, {end=}, {days=}"
        raise YieldWeekdaysError(msg)


class YieldWeekdaysError(Exception):
    ...


__all__ = [
    "AddWeekdaysError",
    "EPOCH_UTC",
    "LocalTimeZoneError",
    "NOW_HKG",
    "NOW_TKY",
    "NOW_UTC",
    "ParseDateError",
    "ParseDateTimeError",
    "ParseTimeError",
    "ParseTimedeltaError",
    "TODAY_HKG",
    "TODAY_TKY",
    "TODAY_UTC",
    "UTC",
    "YieldWeekdaysError",
    "add_weekdays",
    "date_to_datetime",
    "duration_to_float",
    "duration_to_timedelta",
    "ensure_date",
    "ensure_datetime",
    "ensure_time",
    "ensure_timedelta",
    "get_now",
    "get_today",
    "is_weekday",
    "local_timezone",
    "maybe_sub_pct_y",
    "parse_date",
    "parse_datetime",
    "parse_time",
    "parse_timedelta",
    "round_to_next_weekday",
    "round_to_prev_weekday",
    "serialize_date",
    "serialize_datetime",
    "serialize_time",
    "serialize_timedelta",
    "yield_weekdays",
]



from __future__ import annotations

from beartype.roar import BeartypeCallHintParamViolation
from pytest import raises

from tests.beartype.functions import Example, identity


class TestBeartypeIfDev:
    def test_main(self) -> None:
        assert identity(0) == 0
        with raises(BeartypeCallHintParamViolation):
            _ = identity(0.0)  # type: ignore

    def test_dataclass(self) -> None:
        assert Example(0).x == 0
        with raises(BeartypeCallHintParamViolation):
            _ = Example(0.0)  # type: ignore



from __future__ import annotations

from pytest import mark, param, raises

from utilities.errors import RedirectErrorError, redirect_error, retry
from utilities.more_itertools import one


class TestRedirectError:
    def test_redirect(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        with raises(SecondError), redirect_error(FirstError, SecondError):
            raise FirstError

    def test_no_redirect(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        class ThirdError(Exception):
            ...

        with raises(FirstError), redirect_error(SecondError, ThirdError):
            raise FirstError

    def test_match_and_redirect(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        with raises(SecondError), redirect_error(FirstError, SecondError, match="text"):
            msg = "text"
            raise FirstError(msg)

    def test_match_and_args_empty_error(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        with raises(RedirectErrorError), redirect_error(
            FirstError, SecondError, match="match"
        ):
            raise FirstError()  # noqa: RSE102

    def test_match_and_args_non_unique_error(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        with raises(RedirectErrorError), redirect_error(
            FirstError, SecondError, match="match"
        ):
            args = "x", "y"
            raise FirstError(args)

    def test_match_and_arg_not_string_error(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        with raises(RedirectErrorError), redirect_error(
            FirstError, SecondError, match="match"
        ):
            arg = 0
            raise FirstError(arg)

    def test_match_and_no_redirect(self) -> None:
        class FirstError(Exception):
            ...

        class SecondError(Exception):
            ...

        with raises(FirstError), redirect_error(
            FirstError, SecondError, match="something else"
        ):
            msg = "text"
            raise FirstError(msg)


class TestRetry:
    @mark.parametrize("use_predicate", [param(None), param(True), param(False)])
    def test_main(self, *, use_predicate: bool | None) -> None:
        class TooLargeError(Exception):
            ...

        def increment() -> int:
            nonlocal n
            n += 1
            if n >= 3:
                raise TooLargeError(n)
            return n

        n = 0
        assert increment() == 1
        assert increment() == 2
        with raises(TooLargeError):
            _ = increment()

        def reset(_error: TooLargeError, /) -> None:
            nonlocal n
            n = 0

        if use_predicate is None:
            retry_inc = retry(increment, TooLargeError, reset)
        else:

            def predicate(error: TooLargeError, /) -> bool:
                if use_predicate:
                    return one(error.args) >= 3
                return one(error.args) >= 4

            retry_inc = retry(increment, TooLargeError, reset, predicate=predicate)

        n = 0
        assert retry_inc() == 1
        assert retry_inc() == 2
        if (use_predicate is None) or (use_predicate is True):
            assert retry_inc() == 1
        else:
            with raises(TooLargeError):
                _ = retry_inc()



from __future__ import annotations

from re import search

from bidict import ValueDuplicationError, bidict
from humps import decamelize

from utilities.errors import redirect_error
from utilities.iterables import CheckDuplicatesError, check_duplicates
from utilities.types import IterableStrs


def snake_case(text: str, /) -> str:
    """Convert text into snake case."""

    text = decamelize(text)
    while search("__", text):
        text = text.replace("__", "_")
    return text.lower()


def snake_case_mappings(text: IterableStrs, /) -> bidict[str, str]:
    """Map a set of text into their snake cases."""

    text_as_list = list(text)
    with redirect_error(
        CheckDuplicatesError, SnakeCaseMappingsError(f"{text_as_list=}")
    ):
        check_duplicates(text_as_list)

    with redirect_error(
        ValueDuplicationError, SnakeCaseMappingsError(f"{text_as_list=}")
    ):
        return bidict({t: snake_case(t) for t in text_as_list})


class SnakeCaseMappingsError(Exception):
    ...


__all__ = ["SnakeCaseMappingsError", "snake_case", "snake_case_mappings"]



from __future__ import annotations

from typing import Any, TypeVar, cast

from holoviews import Layout, save

from utilities._holoviews.common import apply_opts
from utilities.atomicwrites import writer
from utilities.types import PathLike

_T = TypeVar("_T")


def apply_cols(layout: Layout, ncols: int, /) -> Layout:
    """Apply the `cols` argument to a layout."""
    return layout.cols(ncols)


def relabel_plot(plot: _T, label: str, /) -> _T:
    """Re-label a plot."""
    return cast(Any, plot).relabel(label)


def save_plot(plot: Any, path: PathLike, /, *, overwrite: bool = False) -> None:
    """Atomically save a plot to disk."""
    with writer(path, overwrite=overwrite) as temp:  # pragma: os-ne-linux
        save(plot, temp, backend="bokeh")


__all__ = ["apply_cols", "apply_opts", "relabel_plot", "save_plot"]


try:
    from utilities._holoviews.xarray import PlotCurveError, plot_curve
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["PlotCurveError", "plot_curve"]



from __future__ import annotations

from collections.abc import Callable
from functools import lru_cache
from typing import Any, TypeVar, cast

from cachetools.func import ttl_cache

from utilities.datetime import duration_to_float
from utilities.functions import identity
from utilities.types import Duration

_F = TypeVar("_F", bound=Callable[..., Any])


def cache(
    *, max_size: int | None = None, max_duration: Duration | None = None
) -> Callable[[_F], _F]:
    """Decorate a function with `max_size` and/or `ttl` settings."""

    if max_duration is not None:
        return ttl_cache(maxsize=max_size, ttl=duration_to_float(max_duration))
    if max_size is not None:
        return cast(Any, lru_cache(maxsize=max_size))
    return identity


__all__ = ["cache"]



from __future__ import annotations

from collections.abc import Mapping
from collections.abc import Set as AbstractSet
from pathlib import Path
from re import search
from time import sleep

from hypothesis import given
from hypothesis.strategies import dictionaries, none, sampled_from, sets
from loguru import logger

from utilities.hypothesis import settings_with_reduced_examples, temp_paths, text_ascii
from utilities.logging import LogLevel
from utilities.loguru import (
    _FILES_ENV_VAR,
    _augment_levels,
    _get_files_path,
    setup_loguru,
)
from utilities.os import temp_environ
from utilities.pathvalidate import valid_path
from utilities.types import PathLike


class TestSetupLoguru:
    def test_disabled(self) -> None:
        setup_loguru()
        logger.info("test")

    @given(levels=dictionaries(text_ascii(min_size=1), sampled_from(LogLevel)))
    def test_level(self, *, levels: dict[str, LogLevel]) -> None:
        setup_loguru(levels=levels)
        logger.info("test")

    @given(module=text_ascii(min_size=1))
    def test_standard(self, *, module: str) -> None:
        setup_loguru(levels={module: LogLevel.INFO})
        logger.info("test")

    @given(enable=sets(text_ascii(min_size=1)))
    def test_enable(self, *, enable: AbstractSet[str]) -> None:
        setup_loguru(enable=enable)
        logger.info("test")

    @given(level=sampled_from(LogLevel))
    def test_luigi(self, *, level: LogLevel) -> None:
        setup_loguru(levels={"luigi": level})
        logger.info("test")

    @given(root=temp_paths(), files=text_ascii(min_size=1))
    @settings_with_reduced_examples()
    def test_files(self, *, root: Path, files: str) -> None:
        setup_loguru(files=files, files_root=root)
        sleep(0.05)
        logger.info("message")
        sleep(0.05)
        (files_dir,) = root.iterdir()
        assert files_dir.name == files
        results = {path.name for path in files_dir.iterdir()}
        expected = {"debug", "info", "warning", "error", "log"}
        assert results == expected
        with valid_path(files_dir, "log").open() as fh:
            (line,) = fh.read().splitlines()
        assert search(
            r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3}  INFO  \w+\-\d+  "
            r"[\w\.\-]+  message$",
            line,
        )


env_var_prefixes = text_ascii(min_size=20)
modules = text_ascii(min_size=1).map(str.lower)


class TestAugmentLevels:
    def test_none(self) -> None:
        result = _augment_levels()
        assert result == {}

    @given(levels=dictionaries(text_ascii(min_size=1), sampled_from(LogLevel)))
    def test_main(self, *, levels: Mapping[str, LogLevel] | None) -> None:
        result = _augment_levels(levels=levels)
        assert result == levels

    @given(
        env_var_prefix=env_var_prefixes, module=modules, level=sampled_from(LogLevel)
    )
    def test_with_env_var(
        self, *, env_var_prefix: str, module: str, level: LogLevel
    ) -> None:
        with temp_environ({f"{env_var_prefix}_{module}": level}):
            result = _augment_levels(env_var_prefix=env_var_prefix)
        assert result == {module: level}

    @given(
        env_var_prefix=env_var_prefixes, module=modules, level=sampled_from(LogLevel)
    )
    def test_without_env_var(
        self, *, env_var_prefix: str, module: str, level: LogLevel
    ) -> None:
        with temp_environ({f"{env_var_prefix}_{module}": level}):
            result = _augment_levels(env_var_prefix=None)
        assert result == {}

    @given(
        env_var_prefix=env_var_prefixes,
        module=modules,
        level_direct=sampled_from(LogLevel),
        level_env_var=sampled_from(LogLevel),
    )
    def test_both(
        self,
        *,
        level_direct: LogLevel,
        env_var_prefix: str,
        module: str,
        level_env_var: LogLevel,
    ) -> None:
        with temp_environ({f"{env_var_prefix}_{module}": level_env_var}):
            result = _augment_levels(
                levels={module: level_direct}, env_var_prefix=env_var_prefix
            )
        assert result == {module: level_env_var}


class TestGetFilesPath:
    @given(files=text_ascii(min_size=1) | none())
    def test_main(self, *, files: PathLike | None) -> None:
        result = _get_files_path(files=files)
        assert result == files

    @given(env_var=text_ascii(min_size=1), files=text_ascii(min_size=1))
    def test_with_env_var(self, *, env_var: str, files: str) -> None:
        with temp_environ({env_var: files}):
            result = _get_files_path(env_var=env_var)
        assert result == files

    @given(files=text_ascii(min_size=1))
    def test_without_env_var(self, *, files: str) -> None:
        with temp_environ({_FILES_ENV_VAR: files}):
            result = _get_files_path(env_var=None)
        assert result is None

    @given(
        files=text_ascii(min_size=1),
        env_var_key=env_var_prefixes,
        env_var_value=text_ascii(min_size=1),
    )
    def test_both(self, *, files: str, env_var_key: str, env_var_value: str) -> None:
        with temp_environ({env_var_key: env_var_value}):
            result = _get_files_path(files=files, env_var=env_var_key)
        assert result == files



from __future__ import annotations

from hypothesis import given
from sqlalchemy import Engine

from utilities.hypothesis import sqlite_engines
from utilities.json import deserialize, serialize


class TestSerializeAndDeserialize:
    @given(x=sqlite_engines(), y=sqlite_engines())
    def test_engine(self, *, x: Engine, y: Engine) -> None:
        ser_x = serialize(x)
        assert deserialize(ser_x).url == x.url
        res = ser_x == serialize(y)
        expected = x.url == y.url
        assert res is expected



from __future__ import annotations

import datetime as dt
from itertools import pairwise
from re import search
from subprocess import PIPE, check_output
from typing import Any, cast

from hypothesis import HealthCheck, Phase, assume, given, settings
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    datetimes,
    floats,
    integers,
    just,
    none,
)
from pytest import raises
from semver import Version

from utilities.datetime import UTC
from utilities.git import _GET_BRANCH_NAME
from utilities.hypothesis import (
    assume_does_not_raise,
    datetimes_utc,
    git_repos,
    hashables,
    lists_fixed_length,
    settings_with_reduced_examples,
    setup_hypothesis_profiles,
    slices,
    text_ascii,
    text_clean,
    text_printable,
    versions,
)
from utilities.os import temp_environ
from utilities.pathvalidate import valid_path


class TestAssumeDoesNotRaise:
    @given(x=booleans())
    def test_no_match_and_suppressed(self, *, x: bool) -> None:
        with assume_does_not_raise(ValueError):
            if x is True:
                msg = "x is True"
                raise ValueError(msg)
        assert x is False

    @given(x=booleans())
    def test_no_match_and_not_suppressed(self, *, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(RuntimeError):
                raise ValueError(msg)

    @given(x=booleans())
    def test_with_match_and_suppressed(self, *, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with assume_does_not_raise(ValueError, match=msg):
                raise ValueError(msg)
        assert x is False

    @given(x=just(True))
    def test_with_match_and_not_suppressed(self, *, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                ValueError, match="wrong"
            ):
                raise ValueError(msg)


class TestDatetimesUTC:
    @given(data=data(), min_value=datetimes(), max_value=datetimes())
    def test_main(
        self, *, data: DataObject, min_value: dt.datetime, max_value: dt.datetime
    ) -> None:
        min_value, max_value = (v.replace(tzinfo=UTC) for v in [min_value, max_value])
        _ = assume(min_value <= max_value)
        datetime = data.draw(datetimes_utc(min_value=min_value, max_value=max_value))
        assert min_value <= datetime <= max_value


class TestGitRepos:
    @given(data=data())
    @settings_with_reduced_examples(suppress_health_check={HealthCheck.filter_too_much})
    def test_main(self, *, data: DataObject) -> None:
        branch = data.draw(text_ascii(min_size=1) | none())
        path = data.draw(git_repos(branch=branch))
        assert set(path.iterdir()) == {valid_path(path, ".git")}
        if branch is not None:
            output = check_output(
                _GET_BRANCH_NAME,  # noqa: S603
                stderr=PIPE,
                cwd=path,
                text=True,
            )
            assert output.strip("\n") == branch


class TestHashables:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        x = data.draw(hashables())
        _ = hash(x)


class TestReducedExamples:
    @given(frac=floats(0.0, 10.0))
    def test_main(self, *, frac: float) -> None:
        @settings_with_reduced_examples(frac)
        def test() -> None:
            pass

        result = cast(Any, test)._hypothesis_internal_use_settings.max_examples  # noqa: SLF001
        expected = max(round(frac * settings().max_examples), 1)
        assert result == expected


class TestSlices:
    @given(data=data(), iter_len=integers(0, 10))
    def test_main(self, *, data: DataObject, iter_len: int) -> None:
        slice_len = data.draw(integers(0, iter_len) | none())
        slice_ = data.draw(slices(iter_len, slice_len=slice_len))
        range_slice = range(iter_len)[slice_]
        assert all(i + 1 == j for i, j in pairwise(range_slice))
        if slice_len is not None:
            assert len(range_slice) == slice_len

    @given(data=data(), iter_len=integers(0, 10))
    def test_error(self, *, data: DataObject, iter_len: int) -> None:
        with raises(
            InvalidArgument, match=r"Slice length \d+ exceeds iterable length \d+"
        ):
            _ = data.draw(slices(iter_len, slice_len=iter_len + 1))


class TestSetupHypothesisProfiles:
    def test_main(self) -> None:
        setup_hypothesis_profiles()
        curr = settings()
        assert Phase.shrink in curr.phases
        assert curr.max_examples in {10, 100, 1000}

    def test_no_shrink(self) -> None:
        with temp_environ({"HYPOTHESIS_NO_SHRINK": "1"}):
            setup_hypothesis_profiles()
        assert Phase.shrink not in settings().phases

    @given(max_examples=integers(1, 100))
    def test_max_examples(self, *, max_examples: int) -> None:
        with temp_environ({"HYPOTHESIS_MAX_EXAMPLES": str(max_examples)}):
            setup_hypothesis_profiles()
        assert settings().max_examples == max_examples


class TestTextClean:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        min_size: int,
        max_size: int | None,
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_clean(
                    min_size=min_size, max_size=max_size, disallow_na=disallow_na
                )
            )
        assert search("^\\S[^\\r\\n]*$|^$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"


class TestTextPrintable:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        min_size: int,
        max_size: int | None,
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_printable(
                    min_size=min_size, max_size=max_size, disallow_na=disallow_na
                )
            )
        assert search(r"^[0-9A-Za-z!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~\s]*$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"


class TestVersions:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        version = data.draw(versions())
        assert isinstance(version, Version)

    @given(data=data())
    def test_min_version(self, data: DataObject) -> None:
        min_version = data.draw(versions())
        version = data.draw(versions(min_version=min_version))
        assert version >= min_version

    @given(data=data())
    def test_max_version(self, data: DataObject) -> None:
        max_version = data.draw(versions())
        version = data.draw(versions(max_version=max_version))
        assert version <= max_version

    @given(data=data())
    def test_min_and_max_version(self, data: DataObject) -> None:
        version1, version2 = data.draw(lists_fixed_length(versions(), 2))
        min_version = min(version1, version2)
        max_version = max(version1, version2)
        version = data.draw(versions(min_version=min_version, max_version=max_version))
        assert min_version <= version <= max_version

    @given(data=data())
    def test_error(self, data: DataObject) -> None:
        version1, version2 = data.draw(lists_fixed_length(versions(), 2))
        _ = assume(version1 != version2)
        with raises(InvalidArgument):
            _ = data.draw(
                versions(
                    min_version=max(version1, version2),
                    max_version=min(version1, version2),
                )
            )



from __future__ import annotations

from pytest import raises

from utilities.text import EnsureStrError, ensure_str, strip_and_dedent


class TestEnsureStr:
    def test_str(self) -> None:
        assert isinstance(ensure_str(""), str)

    def test_not_str(self) -> None:
        with raises(EnsureStrError, match="Object .* must be a string; got .* instead"):
            _ = ensure_str(None)


class TestStripAndDedent:
    def test_main(self) -> None:
        text = """
               This is line 1.
               This is line 2.
               """
        result = strip_and_dedent(text)
        expected = "This is line 1.\nThis is line 2."
        assert result == expected



from __future__ import annotations

from collections import Counter
from collections.abc import Hashable, Iterable, Iterator, Mapping, Sized
from collections.abc import Set as AbstractSet
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, TypeGuard, cast

from typing_extensions import Never, assert_never, override

from utilities.errors import ImpossibleCaseError
from utilities.math import is_equal_or_approx
from utilities.more_itertools import one
from utilities.text import ensure_str
from utilities.types import ensure_hashable


def check_duplicates(iterable: Iterable[Hashable], /) -> None:
    """Check if an iterable contains any duplicates."""
    counts = {k: v for k, v in Counter(iterable).items() if v > 1}
    if len(counts) >= 1:
        raise CheckDuplicatesError(iterable=iterable, counts=counts)


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckDuplicatesError(Exception):
    iterable: Iterable[Hashable]
    counts: dict[Hashable, int]

    @override
    def __str__(self) -> str:
        return "Iterable {} must not contain duplicates; got {}".format(
            self.iterable, ", ".join(f"({k}, n={v})" for k, v in self.counts.items())
        )


class _CheckIterablesEqualState(Enum):
    left_longer = auto()
    right_longer = auto()


def check_iterables_equal(left: Iterable[Any], right: Iterable[Any], /) -> None:
    """Check that a pair of iterables are equal."""

    errors: list[tuple[int, Any, Any]] = []
    state: _CheckIterablesEqualState | None
    try:
        for i, (lv, rv) in enumerate(zip(left, right, strict=True)):
            if lv != rv:
                errors.append((i, lv, rv))
    except ValueError as error:
        msg = ensure_str(one(error.args))
        match msg:
            case "zip() argument 2 is longer than argument 1":
                state = _CheckIterablesEqualState.right_longer
            case "zip() argument 2 is shorter than argument 1":
                state = _CheckIterablesEqualState.left_longer
            case _:  # pragma: no cover
                raise ImpossibleCaseError(  # pragma: no cover
                    case=[f"{msg=}"]
                ) from None
    else:
        state = None
    if (len(errors) >= 1) or (state is not None):
        raise CheckIterablesEqualError(
            left=left, right=right, errors=errors, state=state
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckIterablesEqualError(Exception):
    left: Iterable[Any]
    right: Iterable[Any]
    errors: list[tuple[int, Any, Any]]
    state: _CheckIterablesEqualState | None

    @override
    def __str__(self) -> str:
        parts = list(self._yield_parts())
        match parts:
            case (desc,):
                pass
            case first, second:
                desc = "{} and {}".format(first, second)
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return "Iterables {} and {} must be equal; {}".format(
            self.left, self.right, desc
        )

    def _yield_parts(self) -> Iterator[str]:
        if len(self.errors) >= 1:
            error_descs = (
                "({}, {}, i={})".format(lv, rv, i) for i, lv, rv in self.errors
            )
            yield "items were {}".format(", ".join(error_descs))
        match self.state:
            case _CheckIterablesEqualState.left_longer:
                yield "left was longer"
            case _CheckIterablesEqualState.right_longer:
                yield "right was longer"
            case None:
                pass
            case _ as never:  # type: ignore
                assert_never(never)


def check_length(
    obj: Sized,
    /,
    *,
    equal: int | None = None,
    equal_or_approx: int | tuple[int, float] | None = None,
    min: int | None = None,  # noqa: A002
    max: int | None = None,  # noqa: A002
) -> None:
    """Check the length of an object."""
    if (equal is not None) and (len(obj) != equal):
        raise _CheckLengthEqualError(obj=obj, equal=equal)
    if (equal_or_approx is not None) and not is_equal_or_approx(
        len(obj), equal_or_approx
    ):
        raise _CheckLengthEqualOrApproxError(obj=obj, equal_or_approx=equal_or_approx)
    if (min is not None) and (len(obj) < min):
        raise _CheckLengthMinError(obj=obj, min_=min)
    if (max is not None) and (len(obj) > max):
        raise _CheckLengthMaxError(obj=obj, max_=max)


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckLengthError(Exception):
    ...


@dataclass(frozen=True, kw_only=True, slots=True)
class _CheckLengthEqualError(CheckLengthError):
    obj: Sized
    equal: int

    @override
    def __str__(self) -> str:
        return "Object {} must have length {}; got {}".format(
            self.obj, self.equal, len(self.obj)
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class _CheckLengthEqualOrApproxError(CheckLengthError):
    obj: Sized
    equal_or_approx: int | tuple[int, float]

    @override
    def __str__(self) -> str:
        match self.equal_or_approx:
            case target, error:
                desc = "approximate length {} (error {:%})".format(target, error)
            case target:
                desc = "length {}".format(target)
        return "Object {} must have {}; got {}".format(self.obj, desc, len(self.obj))


@dataclass(frozen=True, kw_only=True, slots=True)
class _CheckLengthMinError(CheckLengthError):
    obj: Sized
    min_: int

    @override
    def __str__(self) -> str:
        return "Object {} must have minimum length {}; got {}".format(
            self.obj, self.min_, len(self.obj)
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class _CheckLengthMaxError(CheckLengthError):
    obj: Sized
    max_: int

    @override
    def __str__(self) -> str:
        return "Object {} must have maximum length {}; got {}".format(
            self.obj, self.max_, len(self.obj)
        )


def check_lengths_equal(left: Sized, right: Sized, /) -> None:
    """Check that a pair of sizes objects have equal length."""
    if len(left) != len(right):
        raise CheckLengthsEqualError(left=left, right=right)


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckLengthsEqualError(Exception):
    left: Sized
    right: Sized

    @override
    def __str__(self) -> str:
        return (
            "Sized objects {} and {} must have the same length; got {} and {}".format(
                self.left, self.right, len(self.left), len(self.right)
            )
        )


def check_mappings_equal(left: Mapping[Any, Any], right: Mapping[Any, Any], /) -> None:
    """Check that a pair of mappings are equal."""
    left_keys, right_keys = set(left), set(right)
    try:
        check_sets_equal(left_keys, right_keys)
    except CheckSetsEqualError as error:
        left_extra, right_extra = map(set, [error.left_extra, error.right_extra])
    else:
        left_extra = right_extra = set()
    errors: list[tuple[Any, Any, Any]] = []
    for key in left_keys & right_keys:
        lv, rv = left[key], right[key]
        if lv != rv:
            errors.append((key, lv, rv))
    if (len(left_extra) >= 1) or (len(right_extra) >= 1) or (len(errors) >= 1):
        raise CheckMappingsEqualError(
            left=left,
            right=right,
            left_extra=left_extra,
            right_extra=right_extra,
            errors=errors,
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckMappingsEqualError(Exception):
    left: Mapping[Any, Any]
    right: Mapping[Any, Any]
    left_extra: AbstractSet[Any]
    right_extra: AbstractSet[Any]
    errors: list[tuple[Any, Any, Any]]

    @override
    def __str__(self) -> str:
        parts = list(self._yield_parts())
        match parts:
            case (desc,):
                pass
            case first, second:
                desc = "{} and {}".format(first, second)
            case first, second, third:
                desc = "{}, {} and {}".format(first, second, third)
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return "Mappings {} and {} must be equal; {}".format(
            self.left, self.right, desc
        )

    def _yield_parts(self) -> Iterator[str]:
        if len(self.left_extra) >= 1:
            yield "left had extra keys {}".format(self.left_extra)
        if len(self.right_extra) >= 1:
            yield "right had extra keys {}".format(self.right_extra)
        if len(self.errors) >= 1:
            error_descs = (
                "({}, {}, k={})".format(lv, rv, k) for k, lv, rv in self.errors
            )
            yield "items were {}".format(", ".join(error_descs))


def check_sets_equal(left: AbstractSet[Any], right: AbstractSet[Any], /) -> None:
    """Check that a pair of sets are equal."""
    left_extra = left - right
    right_extra = right - left
    if (len(left_extra) >= 1) or (len(right_extra) >= 1):
        raise CheckSetsEqualError(
            left=left, right=right, left_extra=left_extra, right_extra=right_extra
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckSetsEqualError(Exception):
    left: AbstractSet[Any]
    right: AbstractSet[Any]
    left_extra: AbstractSet[Any]
    right_extra: AbstractSet[Any]

    @override
    def __str__(self) -> str:
        parts = list(self._yield_parts())
        match parts:
            case (desc,):
                pass
            case first, second:
                desc = "{} and {}".format(first, second)
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return "Sets {} and {} must be equal; {}".format(self.left, self.right, desc)

    def _yield_parts(self) -> Iterator[str]:
        if len(self.left_extra) >= 1:
            yield "left had extra items {}".format(self.left_extra)
        if len(self.right_extra) >= 1:
            yield "right had extra items {}".format(self.right_extra)


def check_subset(left: AbstractSet[Any], right: AbstractSet[Any], /) -> None:
    """Check that a set is a subset of another set."""
    extra = left - right
    if len(extra) >= 1:
        raise CheckSubsetError(left=left, right=right, extra=extra)


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckSubsetError(Exception):
    left: AbstractSet[Any]
    right: AbstractSet[Any]
    extra: AbstractSet[Any]

    @override
    def __str__(self) -> str:
        return "Set {} must be a subset of {}; left had extra items {}".format(
            self.left, self.right, self.extra
        )


def check_superset(left: AbstractSet[Any], right: AbstractSet[Any], /) -> None:
    """Check that a set is a superset of another set."""
    extra = right - left
    if len(extra) >= 1:
        raise CheckSupersetError(left=left, right=right, extra=extra)


@dataclass(frozen=True, kw_only=True, slots=True)
class CheckSupersetError(Exception):
    left: AbstractSet[Any]
    right: AbstractSet[Any]
    extra: AbstractSet[Any]

    @override
    def __str__(self) -> str:
        return "Set {} must be a superset of {}; right had extra items {}".format(
            self.left, self.right, self.extra
        )


def ensure_hashables(
    *args: Any, **kwargs: Any
) -> tuple[list[Hashable], dict[str, Hashable]]:
    """Ensure a set of positional & keyword arguments are all hashable."""
    hash_args = list(map(ensure_hashable, args))
    hash_kwargs = {k: ensure_hashable(v) for k, v in kwargs.items()}
    return hash_args, hash_kwargs


def is_iterable_not_str(obj: Any, /) -> TypeGuard[Iterable[Any]]:
    """Check if an object is iterable, but not a string."""
    try:
        iter(obj)
    except TypeError:
        return False
    return not isinstance(obj, str)


__all__ = [
    "CheckDuplicatesError",
    "CheckIterablesEqualError",
    "CheckLengthsEqualError",
    "CheckMappingsEqualError",
    "CheckSetsEqualError",
    "CheckSubsetError",
    "CheckSupersetError",
    "check_duplicates",
    "check_iterables_equal",
    "check_lengths_equal",
    "check_mappings_equal",
    "check_sets_equal",
    "check_subset",
    "check_superset",
    "ensure_hashables",
    "is_iterable_not_str",
]



from __future__ import annotations

from utilities.sys import VERSION_MAJOR_MINOR, is_pytest


class TestIsPytest:
    def test_main(self) -> None:
        assert is_pytest()


class TestVersionMajorMinor:
    def test_main(self) -> None:
        assert isinstance(VERSION_MAJOR_MINOR, tuple)
        expected = 2
        assert len(VERSION_MAJOR_MINOR) == expected



from __future__ import annotations

from collections.abc import Callable
from functools import partial
from inspect import getsource
from operator import le, lt
from re import search
from types import ModuleType
from typing import Any

from pytest import mark, param

import utilities
from tests.modules import package_with, package_without, standalone
from utilities.ast import yield_dunder_all
from utilities.iterables import check_duplicates
from utilities.modules import (
    yield_module_contents,
    yield_module_subclasses,
    yield_modules,
)
from utilities.types import get_class_name


class TestYieldModules:
    @mark.parametrize(
        ("module", "recursive", "expected"),
        [
            param(standalone, False, 1),
            param(standalone, True, 1),
            param(package_without, False, 2),
            param(package_without, True, 2),
            param(package_with, False, 3),
            param(package_with, True, 6),
        ],
    )
    def test_main(self, *, module: ModuleType, recursive: bool, expected: int) -> None:
        res = list(yield_modules(module, recursive=recursive))
        assert len(res) == expected

    def test_all(self) -> None:
        for module in yield_modules(utilities, recursive=True):
            source = getsource(module)
            for dunder_all in yield_dunder_all(source):
                check_duplicates(dunder_all)
                expected = sorted(dunder_all)
                msg = (
                    f"Please paste in\n\t{module.__name__}\nthe following:\n\n\n"
                    f"\t__all__ = {expected}\n\n"
                )
                assert dunder_all == expected, msg


class TestYieldModuleContents:
    @mark.parametrize(
        ("module", "recursive", "factor"),
        [
            param(standalone, False, 1),
            param(standalone, True, 1),
            param(package_without, False, 2),
            param(package_without, True, 2),
            param(package_with, False, 2),
            param(package_with, True, 5),
        ],
    )
    @mark.parametrize(
        ("type_", "predicate", "expected"),
        [
            param(int, None, 3),
            param(float, None, 3),
            param((int, float), None, 6),
            param(type, None, 3),
            param(int, partial(le, 0), 2),
            param(int, partial(lt, 0), 1),
            param(float, partial(le, 0), 2),
            param(float, partial(lt, 0), 1),
        ],
    )
    def test_main(
        self,
        *,
        module: ModuleType,
        type_: type[Any] | tuple[type[Any], ...] | None,
        recursive: bool,
        predicate: Callable[[Any], bool],
        expected: int,
        factor: int,
    ) -> None:
        res = list(
            yield_module_contents(
                module, type=type_, recursive=recursive, predicate=predicate
            )
        )
        assert len(res) == (factor * expected)


class TestYieldModuleSubclasses:
    def predicate(self: Any, /) -> bool:
        return bool(search("1", get_class_name(self)))

    @mark.parametrize(
        ("module", "recursive", "factor"),
        [
            param(standalone, False, 1),
            param(standalone, True, 1),
            param(package_without, False, 2),
            param(package_without, True, 2),
            param(package_with, False, 2),
            param(package_with, True, 5),
        ],
    )
    @mark.parametrize(
        ("type_", "predicate", "expected"),
        [
            param(int, None, 1),
            param(int, predicate, 0),
            param(float, None, 2),
            param(float, predicate, 1),
        ],
    )
    def test_main(
        self,
        *,
        module: ModuleType,
        type_: type[Any],
        recursive: bool,
        predicate: Callable[[type[Any]], bool],
        expected: int,
        factor: int,
    ) -> None:
        it = yield_module_subclasses(
            module, type_, recursive=recursive, predicate=predicate
        )
        assert len(list(it)) == (factor * expected)



from __future__ import annotations

from csv import writer
from pathlib import Path

from click.testing import CliRunner
from pytest import fixture

from utilities.pathvalidate import valid_path
from utilities.platform import IS_WINDOWS
from utilities.scripts.csv_to_markdown import _csv_to_markdown, main
from utilities.text import strip_and_dedent


@fixture()
def path_csv(*, tmp_path: Path) -> Path:
    path = valid_path(tmp_path, "input.csv")
    with path.open(mode="w", newline="" if IS_WINDOWS else None) as fh:
        csv_writer = writer(fh)
        csv_writer.writerow(["Foo", "Bar"])
        csv_writer.writerow(["1,1", "1,2"])
        csv_writer.writerow(["2,1", "2,2"])
    return path


class TestCSVToMarkdown:
    def test_main(self, *, path_csv: Path) -> None:
        result = _csv_to_markdown(path=path_csv)
        expected = """
            |Foo|Bar|
            | :---: | :---: |
            |1,1|1,2|
            |2,1|2,2|
        """
        assert result == strip_and_dedent(expected)

    def test_stdout(self, *, path_csv: Path) -> None:
        runner = CliRunner()
        args = ["--path", str(path_csv)]
        result = runner.invoke(main, args)
        assert result.exit_code == 0



from __future__ import annotations

from enum import Enum, auto
from typing import cast

from hypothesis import given
from hypothesis.strategies import DataObject, data, sampled_from
from pytest import raises

from utilities.enum import ParseEnumError, StrEnum, ensure_enum, parse_enum


class TestParseEnum:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        result = parse_enum(Truth, truth.name)
        assert result is truth

    @given(data=data())
    def test_case_insensitive(self, *, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        name = truth.name
        input_ = data.draw(sampled_from([name, name.upper(), name.lower()]))
        result = parse_enum(Truth, input_, case_sensitive=False)
        assert result is truth

    def test_error_empty(self) -> None:
        class Example(Enum):
            pass

        with raises(ParseEnumError):
            _ = parse_enum(Example, "not-a-member")

    @given(data=data())
    def test_error_non_unique(self, *, data: DataObject) -> None:
        class Example(Enum):
            member = auto()
            MEMBER = auto()

        member = data.draw(sampled_from(Example))
        with raises(ParseEnumError):
            _ = parse_enum(Example, member.name, case_sensitive=False)


class TestEnsureEnum:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        input_ = data.draw(sampled_from([truth, truth.name]))
        result = ensure_enum(Truth, input_)
        assert result is truth


class TestStrEnum:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        class Truth(cast(type[Enum], StrEnum)):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        assert isinstance(truth, Enum)
        assert isinstance(truth, str)
        assert truth == truth.name



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager

import timeout_decorator
from sqlalchemy import Engine, Sequence
from sqlalchemy.exc import DatabaseError
from typing_extensions import assert_never

from utilities._sqlalchemy.common import Dialect, get_dialect
from utilities.errors import redirect_error
from utilities.math import FloatFinNonNeg, IntNonNeg


def next_from_sequence(
    name: str, engine: Engine, /, *, timeout: FloatFinNonNeg | None = None
) -> IntNonNeg | None:
    """Get the next element from a sequence."""

    def inner() -> int:
        seq = Sequence(name)
        try:
            with redirect_next_from_sequence_error(
                engine
            ), engine.begin() as conn:  # pragma: no cover
                return conn.scalar(seq)
        except NextFromSequenceError:
            with engine.begin() as conn:  # pragma: no cover
                _ = seq.create(conn)  # pragma: no cover
            return inner()  # pragma: no cover

    if timeout is None:
        return inner()
    func = timeout_decorator.timeout(seconds=timeout)(inner)  # pragma: no cover
    try:  # pragma: no cover
        return func()  # pragma: no cover
    except timeout_decorator.TimeoutError:  # pragma: no cover
        return None  # pragma: no cover


@contextmanager
def redirect_next_from_sequence_error(engine: Engine, /) -> Iterator[None]:
    """Redirect to the `NextFromSequenceError`."""
    match dialect := get_dialect(engine):
        case (  # pragma: no cover
            Dialect.mssql
            | Dialect.mysql
            | Dialect.postgresql
        ):
            raise NotImplementedError(dialect)  # pragma: no cover
        case Dialect.oracle:  # pragma: no cover
            match = "ORA-02289: sequence does not exist"
        case Dialect.sqlite:
            msg = f"{engine=}"
            raise NotImplementedError(msg)
        case _ as never:  # type: ignore
            assert_never(never)
    with redirect_error(
        DatabaseError, NextFromSequenceError, match=match
    ):  # pragma: no cover
        yield


class NextFromSequenceError(Exception):
    ...


__all__ = [
    "NextFromSequenceError",
    "next_from_sequence",
    "redirect_next_from_sequence_error",
]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from os import chdir
from os import walk as _walk
from pathlib import Path

from utilities.datetime import UTC
from utilities.pathvalidate import valid_path, valid_path_cwd
from utilities.re import extract_group
from utilities.types import PathLike


def ensure_suffix(path: PathLike, suffix: str, /) -> Path:
    """Ensure a path has the required suffix."""
    as_path = valid_path(path)
    parts = as_path.name.split(".")
    clean_suffix = extract_group(r"^\.(\w+)$", suffix)
    if parts[-1] != clean_suffix:
        parts.append(clean_suffix)
    return as_path.with_name(".".join(parts))


def get_modified_time(path: PathLike, /) -> dt.datetime:
    """Get the modified time of a file."""
    return dt.datetime.fromtimestamp(Path(path).stat().st_mtime, tz=UTC)


@contextmanager
def temp_cwd(path: PathLike, /) -> Iterator[None]:
    """Context manager with temporary current working directory set."""
    prev = valid_path_cwd()
    chdir(path)
    try:
        yield
    finally:
        chdir(prev)


def walk(
    top: PathLike,
    /,
    *,
    topdown: bool = True,
    onerror: Callable[[OSError], None] | None = None,
    followlinks: bool = False,
) -> Iterator[tuple[Path, list[Path], list[Path]]]:
    """Iterate through a directory recursively."""
    for dirpath, dirnames, filenames in _walk(
        top, topdown=topdown, onerror=onerror, followlinks=followlinks
    ):
        yield (
            valid_path(dirpath),
            list(map(valid_path, dirnames)),
            list(map(valid_path, filenames)),
        )


__all__ = ["ensure_suffix", "get_modified_time", "temp_cwd", "walk"]
